import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:build/build.dart';
import 'package:dartxx/dartxx.dart';
import 'package:flutter_degen/flutter_degen.dart';
import 'package:source_gen/source_gen.dart';
import 'package:sunny_fluro/annotations.dart';
import 'package:sunny_fluro_generator/fluro_generator.dart';
import 'package:sunny_fluro_generator/read_metadata.dart';

class RoutesGenerator extends GeneratorForAnnotation<routes> {
  @override
  dynamic generateForAnnotatedElement(Element element, ConstantReader annotation, BuildStep buildStep) {
    assert(element is ClassElement, "@routes must only be applied to class");
    final cls = element as ClassElement;
    assert(cls.isAbstract, "@routes must be applied to abstract classes");
    List<String> mixin = [];

    /// First letter is underscore
    final concreteName = element.name.substring(1);
    // ignore: unused_local_variable
    final fieldType = cls;

    final fieldsToProcess = cls.fields.mapNotNull((fld) {
      final metasrc = fld.isSynthetic ? fld.getter : fld;
      final annotated = metasrc!.readAnnotation<route>();
      return annotated?.hasAnnotation == true ? annotated : null;
    });

    mixin += [
      "class $concreteName extends ${element.name} {",
      "",
      for (var fld in fieldsToProcess) ...[
        "  @override",
        "  final ${fieldName(fld.element)!.capitalize()}Route ${fieldName(fld.element)};",
        "",
      ],
      "",
      "  $concreteName(FRouter router): ",
      fieldsToProcess
          .map((fld) => "    ${fieldName(fld.element)} =  router.register(${fieldName(fld.element)!.capitalize()}Route(router))")
          .join(",\n"),
      "  ;",
    ];

    mixin += [
      // "  ${fieldType.name} get ${field.name};",
      "}",
    ];

    fieldsToProcess.forEach((fld) {
      /// First letter is underscore
      final concreteName = "${fld.elementName.capitalize()}Route";

      final parsed = RouteDefinition.from(fld.reader!);
      if (parsed.widgetType != null) {
        final widgetClass = parsed.widgetType!.element as ClassElement;
        ConstructorElement ctor = widgetClass.constructors.firstWhere((element) => element.displayName == parsed.ctor);

        /// This is a widget route
        mixin += [
          """
class $concreteName extends UriTemplateAppPageRoute<${parsed.returns.safeName}, ${concreteName}Params> {
  final FRouter router;
  $concreteName(this.router)
    : super.ofUri(
        "${parsed.uri}",
        (context, ${concreteName}Params? p) => ${invokeDelegate(
            ctor.parameters,
            delegateName: "${parsed.widgetType!.safeName}${ctor.displayName.isNullOrBlank ? '' : '.${ctor.displayName}'}",
            delegatePrefix: 'p!.',
          )},
        (dyn) => ${concreteName}Params.of(dyn),
        name: "${parsed.name}",
      );
              
  Future call(BuildContext context, {
    ${functionParams(ctor.parameters)}
    bool replace = false,
    bool clearStack = false,
    TransitionType? transition,
    bool rootNavigator = false,
    NavigatorState? navigator,
    Duration transitionDuration = const Duration(milliseconds: 250),
    RouteTransitionsBuilder? transitionBuilder}) {
      return this.router.navigateToRoute(
        context,
        this,
        parameters: ${concreteName}Params(
        ${paramPass(ctor.parameters)}
        ),
        rootNavigator: rootNavigator,
        clearStack: clearStack,
        navigator: navigator,
        replace: replace,
        transition: transition,
        transitionBuilder: transitionBuilder,
        transitionDuration: transitionDuration,);
  }
}
          """,
        ];

        mixin += [
          "",
          "",
          "class ${concreteName}Params extends DefaultRouteParams {",
          ...paramsClassBody(concreteName, ctor.parameters),
          "}",
        ];
      } else {
        /// This is a completable route

        /// This is a widget route
        var ft = (parsed.handler?.type as FunctionType?);
        if (ft == null) return null;
        var fel = parsed.handler!.toFunctionValue();

        mixin += [
          """
class $concreteName extends UriTemplateCompletableAppRoute<${parsed.returns.safeName}?, ${concreteName}Params> {
  final FRouter router;
  $concreteName(this.router): super.ofUri(
    "${parsed.uri ?? "/${fld.elementName.uncapitalize()}"}",
    (context, ${concreteName}Params p, sendRoute) => ${delegateFunction(ft, fel!.displayName, asExpression: true, sourcePrefix: (param) => param.type.safeName == "BuildContext" ? '' : 'p.')},
    (dyn) => ${concreteName}Params.of(dyn),
    name: "${parsed.name}",
  );
    
  Future call(BuildContext context, {
    ${functionParams(ft.parameters)}
    bool replace = false,
    bool clearStack = false,
    TransitionType? transition,
    bool rootNavigator = false,
    NavigatorState? navigator,
    Duration transitionDuration = const Duration(milliseconds: 250),
    RouteTransitionsBuilder? transitionBuilder}) {
  return this.router.navigateToRoute(
    context,
    this,
    parameters: ${concreteName}Params(
    ${paramPass(ft.parameters)}
    ),
    rootNavigator: rootNavigator,
    clearStack: clearStack,
    navigator: navigator,
    replace: replace,
    transition: transition,
    transitionBuilder: transitionBuilder,
    transitionDuration: transitionDuration,);
  }
}  
    
class ${concreteName}Params extends DefaultRouteParams {""",
          ...paramsClassBody(concreteName, ft.parameters),
          '}',
          '',
        ];
      }
    });

    return mixin.join("\n");
  }

  String thisParams(List<ParameterElement> parameters) {
    String str = '';
    for (var param in parameters) {
      if (param.type.safeName != 'BuildContext') {
        str +=
            '    ${param.isRequiredPositional || param.isRequiredNamed ? 'required ' : ''}this.${param.name}${param.hasDefaultValue ? ' = ${param.defaultValueCode}' : ''},\n';
      }
    }
    return str;
  }

  String functionParams(List<ParameterElement> parameters) {
    String str = '';
    for (var param in parameters) {
      if (param.type.safeName != 'BuildContext') {
        str +=
            '    ${param.isRequiredPositional || param.isRequiredNamed ? 'required ' : ''} ${param.type.nullSafeName} ${param.name}${param.hasDefaultValue ? ' = ${param.defaultValueCode}' : ''},\n';
      }
    }
    return str;
  }

  String paramPass(List<ParameterElement> parameters) {
    String str = '';
    for (var param in parameters) {
      if (param.type.safeName != 'BuildContext') {
        str += '    ${param.name}: ${param.name},\n';
      }
    }
    return str;
  }

  List<String> paramsClassBody(String concreteName, List<ParameterElement> params) {
    params = params.where((element) => element.type.safeName != "BuildContext").toList();
    return [
      for (var param in params)
        if (param.type.safeName != "BuildContext") '  final ${param.type.nullSafeName} ${param.displayName};',
      'factory ${concreteName}Params.of(dynamic) {',
      '  if (dynamic is ${concreteName}Params) {',
      '    return dynamic;',
      '  }',
      '  return ${concreteName}Params(',
      for (var param in params)
        if (param.type.safeName != "BuildContext") '      ${param.name}: dynamic["${param.name}"] as ${param.type.nullSafeName},',
      '  );',
      '}',
      '',
      '',
      '${concreteName}Params(${params.isNotEmpty ? '{' : ''}',
      for (var param in params)
        if (param.type.safeName != "BuildContext")
          '    ${param.isRequiredPositional || param.isRequiredNamed ? 'required ' : ''}this.${param.name}${param.hasDefaultValue ? ' = ${param.defaultValueCode}' : ''},',
      '  ${params.isNotEmpty ? '}' : ''});',
    ];
  }

  String delegateMethod(MethodElement method, String delegateName) {
    var str = "  ${method.returnType.getDisplayString(withNullability: false)} ${method.name}(";
    bool hitPositional = false;
    for (final p in method.parameters) {
      if (p.isNamed && !hitPositional) {
        str += "{";
        hitPositional = true;
      } else if (!p.isNamed && hitPositional) {
        hitPositional = false;
        str += "},";
      }
      str += "${p.type.getDisplayString(withNullability: false)} ${p.name},";
    }
    if (hitPositional) {
      str += "}";
    } else {
      str = str.substring(0, str.length - 1);
    }

    str += ") => ${delegateName}.${method.name}(";
    for (final p in method.parameters) {
      if (p.isNamed) {
        str += "${p.name}: ${p.name},";
      } else {
        str += "${p.name},";
      }
    }
    str = str.substring(0, str.length - 1);
    str += ");";
    return str;
  }

  String invokeDelegate(List<ParameterElement> parameters, {required String delegateName, String delegatePrefix = ''}) {
    var str = "${delegateName}(";

    for (final p in parameters) {
      if (p.isNamed) {
        str += "${p.name}: ${delegatePrefix}${p.name},";
      } else {
        str += "${delegatePrefix}${p.name},";
      }
    }
    str += ")";
    return str;
  }
}

extension AnnotatedNameExt<T> on Annotated<T> {
  String get elementName {
    return fieldName(element)!;
  }
}

class RouteDefinition {
  final String? uri;
  final DartType? widgetType;
  final DartType? returns;
  final String? ctor;
  final DartObject? handler;
  final DartType? params;
  final String? name;

  const RouteDefinition({
    required this.uri,
    this.widgetType,
    required this.returns,
    this.ctor,
    required this.handler,
    required this.params,
    required this.name,
  });

  factory RouteDefinition.from(ConstantReader annotation) {
    return RouteDefinition(
        uri: annotation.get("uri"),
        widgetType: annotation.get("widgetType"),
        returns: annotation.get("returns"),
        ctor: annotation.get("ctor"),
        handler: annotation.get("handler"),
        params: annotation.get("params"),
        name: annotation.get("name"));
  }
}

extension ConstantReaderNullSafeExt on ConstantReader {
  T get<T>(String field) {
    final v = read(field);
    if (v.isNull) {
      return null as T;
    } else if (v.isLiteral) {
      return v.literalValue as T;
    } else if (v.isList) {
      return v.listValue as T;
    } else if (v.isType) {
      return v.typeValue as T;
    } else if (v.isSet) {
      return v.setValue as T;
    } else if (v.isNull) {
      return null as T;
    } else {
      return v.objectValue as T;
    }
  }
}

extension DartTypeName on DartType? {
  String get safeName {
    return this?.getDisplayString(withNullability: false) ?? 'dynamic';
  }

  String get nullSafeName {
    return this?.getDisplayString(withNullability: true) ?? 'dynamic';
  }
}

typedef SourcePrefixGenerator = String Function(ParameterElement element);
String defaultPrefix(ParameterElement element) => '';
String delegateFunction(FunctionType method, String delegateName,
    {bool asExpression = false, SourcePrefixGenerator sourcePrefix = defaultPrefix}) {
  var str = "$delegateName(";
  for (final p in method.parameters) {
    str += "${p.pass(sourcePrefix: sourcePrefix(p))}, ";
  }
  str = str.substring(0, str.length - 1);
  str += ")${asExpression ? '' : ';'}";
  return str;
}

extension ParameterPassthroughWithPrefix on ParameterElement {
  String pass({String sourcePrefix = ''}) {
    var str = "";
    var p = this;
    if (p.isNamed) {
      str += "${p.name}: ${sourcePrefix}${p.name}";
    } else {
      str += "${sourcePrefix}${p.name}";
    }
    return str;
  }
}

String delegateFunctionDef(FunctionType method) {
  var str = "";
  bool hitPositional = false;
  for (final p in method.parameters) {
    if (p.isNamed && !hitPositional) {
      str += "{";
      hitPositional = true;
    } else if (!p.isNamed && hitPositional) {
      hitPositional = false;
      str += "},";
    }
    str += "${p.redeclare},";
  }
  if (hitPositional) {
    str += "}";
  } else {
    str = str.substring(0, str.length - 1);
  }
  return str;
}
